
# pyright: reportIncompatibleMethodOverride=false

# Imports
from __future__ import annotations

import random
from collections.abc import Callable, Iterable
from enum import Enum
from typing import Any, Literal, NamedTuple

import cv2
from numpy.typing import NDArray
from PIL import Image

from ..range_tuple import RangeTuple

# Import folders
from .image import (
	adaptive_denoise_image,
	auto_contrast_image,
	bias_field_correction_image,
	bilateral_denoise_image,
	binary_threshold_image,
	blur_image,
	brightness_image,
	canny_image,
	clahe_image,
	contrast_image,
	curvature_flow_filter_image,
	flip_image,
	invert_image,
	laplacian_image,
	median_blur_image,
	nlm_denoise_image,
	noise_image,
	normalize_image,
	random_erase_image,
	resize_image,
	rotate_image,
	salt_pepper_image,
	sharpen_image,
	shear_image,
	threshold_image,
	translate_image,
	tv_denoise_image,
	wavelet_denoise_image,
	zoom_image,
)


# Tuple class
class ProcessingTechnique(NamedTuple):
	""" A named tuple containing an processing technique.

	The following descriptions are the recommendations:

	- rotation:			Rotation between -20° and +20° is generally safe
	- translation:		Translation up to 15% to avoid position bias
	- shearing:			Shearing up to 15° on x and y axes
	- noise:			Gaussian noise with fixed values [0.3,0.4,0.5]
	- salt-pepper:		Salt-pepper noise with densities [0.01,0.02,0.03]
	- sharpening:		Gaussian blur (variance=1) then subtract from original
	- contrast:			Linear scaling between min/max intensities [0.7,1.3]
	- axis_flip:		Flip along x and/or y axis (0 for x, 1 for y, 2 for both), probability=0.5
	- zoom:				Magnification/zoom between 85% and 115%
	- brightness:		Adjust overall image brightness [0.7,1.3]
	- blur:				Gaussian blur for reduced sharpness [0.5,2.0]
	- random_erase:		Randomly erase rectangles (1-10% of image) to learn descriptive features
	- clahe:			Contrast Limited Adaptive Histogram Equalization (clip_limit=[1.0,4.0], tile_size=8)
	- binary_threshold:	Binary thresholding with threshold [0.1,0.9]
	- threshold:		Multi-level thresholding with levels [0.3,0.6]
	- canny:            Canny edge detection (thresholds=[50/255,150/255], aperture=3)
	- laplacian:        Laplacian edge detection with 3x3 kernel
	- auto_contrast:    Auto contrast the image
	- bias_field_correction: Bias field correction
	- curvature_flow_filter: Curvature flow filter (time_step=0.05, iterations=5)
	- median_blur:      Median blur (kernel_size=15, iterations=10)
	- resize:			Resize the image (width=224, height=224, resample=Image.Resampling.LANCZOS)
	- normalize:        Normalize the image (min=0, max=255, method=cv2.NORM_MINMAX)
	- nlm_denoise:      Non-local means denoising (h=10, template_window_size=7, search_window_size=21)
	- bilateral_denoise: Bilateral filter denoising (d=9, sigma_color=75, sigma_space=75)
	- tv_denoise:       Total variation denoising (weight=0.1, iterations=30)
	- wavelet_denoise:  Wavelet denoising (wavelet='db1', mode='soft', wavelet_levels=3)
	- adaptive_denoise: Adaptive denoising (method="nlm", strength=0.5)
	- custom:			Custom processing technique (callable) # Ex: ProcessingTechnique("custom", custom=f)
	"""
	# The following descriptions are the recommendations
	name: Literal[
		"rotation", "translation", "shearing", "noise", "salt-pepper", "sharpening",
		"contrast", "axis_flip", "zoom", "brightness", "blur", "random_erase", "clahe",
		"binary_threshold", "threshold", "canny", "laplacian", "auto_contrast",
		"bias_field_correction", "curvature_flow_filter", "resize",
		"normalize", "median_blur", "nlm_denoise", "bilateral_denoise",
		"tv_denoise", "wavelet_denoise", "adaptive_denoise", "invert", "custom"
	]
	""" Name of the processing technique """
	ranges: list[Iterable[Any] | RangeTuple]
	""" List of ranges for the processing technique.
	Depending on the technique, multiple ranges might be needed.
	Ex: Translation (x and y axes) or Shearing (x and y axes). """
	probability: float = 1.0
	""" Probability of applying the processing technique (default: 1.0).
	Should be used on techniques like "axis_flip" or "random_erase"
	where the probability of applying the technique is not 100%. """
	custom: Callable[..., NDArray[Any]] | None = None
	""" Custom processing technique (callable), name must be "custom", e.g. ProcessingTechnique("custom", custom=f) """

	def __str__(self) -> str:
		return (
			f"name={self.name}, ranges={self.ranges}, probability={self.probability}, "
			f"custom={self.custom.__name__ if self.custom else None}"
		)

	def __repr__(self) -> str:
		return (
			f"ProcessingTechnique(name={self.name!r}, ranges={self.ranges!r}, "
			f"probability={self.probability}, custom={self.custom.__name__ if self.custom else None}"
		)

	def __mul__(self, other: float) -> ProcessingTechnique:
		""" Multiply all ranges by a scalar value.

		Args:
			other (float): Value to multiply ranges by
		Returns:
			ProcessingTechnique: New object with scaled ranges
		"""
		return ProcessingTechnique(
			self.name,
			[
				(r * other) if isinstance(r, RangeTuple) else [x * other if isinstance(x, float | int) else x for x in r]
				for r in self.ranges
			],
			probability=self.probability,
			custom=self.custom
		)

	def __truediv__(self, other: float) -> ProcessingTechnique:
		""" Divide all ranges by a scalar value.

		Args:
			other (float): Value to divide ranges by
		Returns:
			ProcessingTechnique: New object with scaled ranges
		"""
		return ProcessingTechnique(
			self.name,
			[
				(r / other) if isinstance(r, RangeTuple) else [x / other if isinstance(x, float | int) else x for x in r]
				for r in self.ranges
			],
			probability=self.probability,
			custom=self.custom
		)

	def deterministic(self, use_default: bool = False) -> ProcessingTechnique:
		""" Convert the RangeTuple to values by calling the RangeTuple.random() method. """
		# Make values deterministic
		values: list[Iterable[Any]] = []
		for range in self.ranges:
			if isinstance(range, RangeTuple):
				if use_default:
					values.append([range.default])
				else:
					values.append([range.random()])
			else:
				values.append(range)

		# Make probability deterministic (0 or 1)
		probability: float = 1.0 if random.random() < self.probability else 0.0
		return ProcessingTechnique(self.name, values, probability=probability, custom=self.custom)

	def apply(self, image: NDArray[Any], dividers: tuple[float, float] = (1.0, 1.0), times: int = 1) -> NDArray[Any]:
		""" Apply the processing technique to the image.

		Args:
			image     (NDArray[Any]):           Image to apply the processing technique to
			dividers  (tuple[float, float]):    Dividers used to adjust the processing technique parameters (default: (1.0, 1.0))
			times     (int):                    Number of times to apply the processing technique (default: 1)
		Returns:
			NDArray[Any]: Processed image
		"""
		assert not any(isinstance(x, RangeTuple) for x in self.ranges), (
			"All RangeTuples must be converted to values, "
			f"please call deterministic() before. {self.ranges=}"
		)

		# Check if the technique is applied
		if random.random() > self.probability:
			return image

		for _ in range(times):

			# Apply the processing technique
			if self.name == "rotation":
				angle: float = next(iter(self.ranges[0]))
				image = rotate_image(image, angle)

			elif self.name == "translation":
				x_shift: float = next(iter(self.ranges[0])) / dividers[0]
				y_shift: float = next(iter(self.ranges[1])) / dividers[1]
				image = translate_image(image, x_shift, y_shift)

			elif self.name == "shearing":
				x_shear: float = next(iter(self.ranges[0])) / dividers[0]
				y_shear: float = next(iter(self.ranges[1])) / dividers[1]
				image = shear_image(image, x_shear, y_shear)

			elif self.name == "axis_flip":
				axis: Literal["horizontal", "vertical", "both"] = next(iter(self.ranges[0]))
				image = flip_image(image, axis)

			elif self.name == "noise":
				intensity: float = next(iter(self.ranges[0]))
				image = noise_image(image, intensity)

			elif self.name == "salt-pepper":
				density: float = next(iter(self.ranges[0]))
				image = salt_pepper_image(image, density)

			elif self.name == "sharpening":
				alpha: float = next(iter(self.ranges[0]))
				image = sharpen_image(image, alpha)

			elif self.name == "contrast":
				factor: float = next(iter(self.ranges[0]))
				image = contrast_image(image, factor)

			elif self.name == "zoom":
				zoom_factor: float = next(iter(self.ranges[0])) / max(dividers)
				image = zoom_image(image, zoom_factor)

			elif self.name == "brightness":
				brightness_factor: float = next(iter(self.ranges[0]))
				image = brightness_image(image, brightness_factor)

			elif self.name == "blur":
				sigma: float = next(iter(self.ranges[0]))
				image = blur_image(image, sigma)

			elif self.name == "random_erase":
				ratio: float = next(iter(self.ranges[0])) / max(dividers)
				image = random_erase_image(image, ratio)

			elif self.name == "clahe":
				clip_limit: float = next(iter(self.ranges[0]))
				tile_grid_size: int = int(next(iter(self.ranges[0])))
				image = clahe_image(image, clip_limit, tile_grid_size)

			elif self.name == "binary_threshold":
				threshold: float = next(iter(self.ranges[0]))
				image = binary_threshold_image(image, threshold)

			elif self.name == "threshold":
				thresholds: list[float] = [next(iter(r)) for r in self.ranges]
				image = threshold_image(image, thresholds)

			elif self.name == "canny":
				threshold1: float = next(iter(self.ranges[0]))
				threshold2: float = next(iter(self.ranges[1]))
				aperture_size: int = int(next(iter(self.ranges[2]))) if len(self.ranges) > 2 else 3
				image = canny_image(image, threshold1, threshold2, aperture_size)

			elif self.name == "laplacian":
				kernel_size: int = int(next(iter(self.ranges[0]))) if self.ranges else 3
				image = laplacian_image(image, kernel_size)

			elif self.name == "auto_contrast":
				image = auto_contrast_image(image)

			elif self.name == "curvature_flow_filter":
				time_step: float = next(iter(self.ranges[0]))
				number_of_iterations: int = int(next(iter(self.ranges[1])))
				image = curvature_flow_filter_image(image, time_step, number_of_iterations)

			elif self.name == "bias_field_correction":
				image = bias_field_correction_image(image)

			elif self.name == "resize":
				width: int = int(next(iter(self.ranges[0])))
				height: int = int(next(iter(self.ranges[1])))
				if len(self.ranges) > 2:
					image = resize_image(image, width, height, Image.Resampling(next(iter(self.ranges[2]))))
				else:
					image = resize_image(image, width, height)

			elif self.name == "normalize":
				mini: float | int = next(iter(self.ranges[0]))
				maxi: float | int = next(iter(self.ranges[1]))
				norm_method: int = int(next(iter(self.ranges[2])))
				image = normalize_image(image, mini, maxi, norm_method)

			elif self.name == "median_blur":
				kernel_size: int = int(next(iter(self.ranges[0])))
				iterations: int = int(next(iter(self.ranges[1])))
				image = median_blur_image(image, kernel_size, iterations)

			elif self.name == "nlm_denoise":
				h: float = float(next(iter(self.ranges[0])))
				template_window_size: int = int(next(iter(self.ranges[1]))) if len(self.ranges) > 1 else 7
				search_window_size: int = int(next(iter(self.ranges[2]))) if len(self.ranges) > 2 else 21
				image = nlm_denoise_image(image, h, template_window_size, search_window_size)

			elif self.name == "bilateral_denoise":
				d: int = int(next(iter(self.ranges[0])))
				sigma_color: float = float(next(iter(self.ranges[1]))) if len(self.ranges) > 1 else 75.0
				sigma_space: float = float(next(iter(self.ranges[2]))) if len(self.ranges) > 2 else 75.0
				image = bilateral_denoise_image(image, d, sigma_color, sigma_space)

			elif self.name == "tv_denoise":
				weight: float = float(next(iter(self.ranges[0])))
				iterations: int = int(next(iter(self.ranges[1]))) if len(self.ranges) > 1 else 30
				method_value = str(next(iter(self.ranges[2]))) if len(self.ranges) > 2 else "chambolle"
				tv_method: Literal["chambolle", "bregman"] = "chambolle" if method_value == "chambolle" else "bregman"
				image = tv_denoise_image(image, weight, iterations, tv_method)

			elif self.name == "wavelet_denoise":
				wavelet_levels: int = int(next(iter(self.ranges[0])))
				wavelet_value = str(next(iter(self.ranges[1]))) if len(self.ranges) > 1 else "db1"
				mode_value = str(next(iter(self.ranges[2]))) if len(self.ranges) > 2 else "soft"
				# sigma is None by default in the function, so we don't provide it explicitly
				image = wavelet_denoise_image(
					image,
					wavelet=wavelet_value,
					mode=mode_value,
					wavelet_levels=wavelet_levels
				)

			elif self.name == "adaptive_denoise":
				method_value = str(next(iter(self.ranges[0])))
				if len(self.ranges) > 1:
					image = adaptive_denoise_image(image, method_value, float(next(iter(self.ranges[1]))))
				else:
					image = adaptive_denoise_image(image, method_value)

			elif self.name == "invert":
				image = invert_image(image)

			elif self.name == "custom":
				if self.custom is None:
					raise ValueError(
						"Custom processing technique is not defined, please set the custom attribute, "
						"ex: ProcessingTechnique('custom', custom=f)"
					)
				args: list[Any] = [next(iter(r)) for r in self.ranges]
				image = self.custom(image, *args)

			else:
				raise ValueError(f"Augmentation technique {self.name} is not supported.")

		return image

	def __call__(self, image: NDArray[Any], dividers: tuple[float, float] = (1.0, 1.0), times: int = 1) -> NDArray[Any]:
		""" Apply the processing technique to the image.

		Args:
			image		(NDArray[Any]):			Image to apply the processing technique to
			dividers	(tuple[float, float]):	Dividers used to adjust the processing technique parameters
				(default: (1.0, 1.0))
			times		(int):					Number of times to apply the processing technique
				(default: 1)
		Returns:
			NDArray[Any]: Processed image
		"""
		return self.apply(image, dividers, times)


# Recommendations enumerated
class RecommendedProcessingTechnique(Enum):
	""" A class containing the processing techniques with their recommended ranges based on scientific papers. """

	ROTATION = ProcessingTechnique("rotation", [RangeTuple(mini=-20, maxi=20, step=1, default=0)])
	""" Rotation between -20° and +20° is generally safe for medical images """

	TRANSLATION = ProcessingTechnique("translation", list(2 * [RangeTuple(mini=-0.15, maxi=0.15, step=0.01, default=0)]))
	""" Translation between -15% and 15% of image size """

	SHEARING = ProcessingTechnique("shearing", list(2 * [RangeTuple(mini=-10, maxi=10, step=1, default=0)]))
	""" Shearing between -10% and 10% distortion """

	AXIS_FLIP = ProcessingTechnique("axis_flip", [["horizontal"]], probability=0.5)
	""" Flipping: horizontal is much more used than vertical """

	NOISE = ProcessingTechnique("noise", [RangeTuple(mini=0.1, maxi=0.4, step=0.05, default=0.2)])
	""" Noise: FV values between 0.1 and 0.4 for gaussian noise """

	SALT_PEPPER = ProcessingTechnique("salt-pepper", [RangeTuple(mini=0.1, maxi=0.5, step=0.05, default=0.2)])
	""" Salt-pepper: densities between 0.1 and 0.5 """

	SHARPENING = ProcessingTechnique("sharpening", [RangeTuple(mini=0.5, maxi=1.5, step=0.1, default=1.1)])
	""" Sharpening: gaussian blur with variance=1 then subtract from original """

	CONTRAST = ProcessingTechnique("contrast", [RangeTuple(mini=0.7, maxi=1.3, step=0.1, default=1.1)])
	""" Contrast: linear scaling between min and max intensity """

	ZOOM = ProcessingTechnique("zoom", [RangeTuple(mini=0.85, maxi=1.15, step=0.05, default=1.05)])
	""" Zoom (Magnification): between 85% and 115% """

	BRIGHTNESS = ProcessingTechnique("brightness", [RangeTuple(mini=0.7, maxi=1.3, step=0.1, default=1.1)])
	""" Brightness: moderate changes to avoid losing details """

	BLUR = ProcessingTechnique("blur", [RangeTuple(mini=0.5, maxi=2, step=0.25, default=1)])
	""" Blur: gaussian blur preferred, moderate values """

	RANDOM_ERASE = ProcessingTechnique("random_erase", [RangeTuple(mini=0.01, maxi=0.1, step=0.01, default=0.05)])
	""" Random erasing: small regions to force model to learn descriptive features """

	CLAHE = ProcessingTechnique(
		"clahe",
		[
			RangeTuple(mini=1.0, maxi=4.0, step=0.5, default=2.0),  # clip_limit
			[8]  # tile_grid_size (fixed at 8x8)
		]
	)
	""" CLAHE: Contrast Limited Adaptive Histogram Equalization with recommended parameters """

	BINARY_THRESHOLD = ProcessingTechnique("binary_threshold", [RangeTuple(mini=0.1, maxi=0.9, step=0.1, default=0.5)])
	""" Binary threshold: threshold between 0.1 and 0.9 """

	THRESHOLD = ProcessingTechnique("threshold", [[0.3], [0.6]])
	""" Multi-level threshold: creates X levels of thresholding """

	CANNY = ProcessingTechnique("canny", [[50 / 255], [150 / 255], [3]])
	""" Canny edge detection with recommended parameters """

	LAPLACIAN = ProcessingTechnique("laplacian", [[3]])  # kernel_size
	""" Laplacian edge detection with 3x3 kernel """

	AUTO_CONTRAST = ProcessingTechnique("auto_contrast", [])
	""" Auto contrast the image """

	CURVATURE_FLOW_FILTER = ProcessingTechnique("curvature_flow_filter", [[0.05], [5]])
	""" Curvature flow filter with recommended parameters """

	BIAS_FIELD_CORRECTION = ProcessingTechnique("bias_field_correction", [])
	""" Bias field correction with recommended parameters """

	NORMALIZE = ProcessingTechnique("normalize", [[0], [255], [cv2.NORM_MINMAX]])
	""" Normalize the image to the range 0-255 """

	MEDIAN_BLUR = ProcessingTechnique("median_blur", [[7], [1]])
	""" Median blur with 7x7 kernel and 1 iteration """

	NLM_DENOISE = ProcessingTechnique("nlm_denoise", [
		RangeTuple(mini=5, maxi=20, step=1, default=10),  # h: affects the strength of the denoising
		[7],                                              # template_window_size: size of the template window
		[21]                                              # search_window_size: size of the search window
	])
	""" Non-local means denoising with recommended parameters """

	BILATERAL_DENOISE = ProcessingTechnique("bilateral_denoise", [
		[9],                                                 # diameter: size of the pixel neighborhood
		RangeTuple(mini=30, maxi=150, step=10, default=75),  # sigma_color: filter sigma in the color space
		RangeTuple(mini=30, maxi=150, step=10, default=75)   # sigma_space: filter sigma in the coordinate space
	])
	""" Bilateral filter denoising with recommended parameters """

	TV_DENOISE = ProcessingTechnique("tv_denoise", [
		RangeTuple(mini=0.05, maxi=0.5, step=0.05, default=0.1),   # weight: denoising weight
		[30],                                                      # max_iter: maximum number of iterations
		["chambolle"]                                              # method: algorithm used for denoising
	])
	""" Total variation denoising with recommended parameters """

	WAVELET_DENOISE = ProcessingTechnique("wavelet_denoise", [
		[3],        # wavelet_levels: number of wavelet decomposition levels
		["db1"],    # wavelet: wavelet to use
		["soft"]    # mode: thresholding mode
	])
	""" Wavelet denoising with recommended parameters """

	ADAPTIVE_DENOISE = ProcessingTechnique("adaptive_denoise", [
		["nlm"],                                               # method: denoising method to use
		RangeTuple(mini=0.1, maxi=0.9, step=0.1, default=0.5)  # strength: denoising strength parameter
	])
	""" Adaptive denoising with recommended parameters """

	INVERT = ProcessingTechnique("invert", [], probability=0.5)
	""" Invert the colors of an image with a 50% probability, hoping the model doesn't focus on bright parts only """




""" Downloader module for the installer subpackage.

Provides functions for downloading and installing programs from URLs.
It handles platform-specific downloads, checking if programs are already installed,
and setting up the downloaded programs for use.

This module works with the main installer module to provide a complete installation
solution for programs that need to be downloaded from the internet.
"""
# Imports
import os
import platform
import subprocess
import sys

from ..print import info, warning
from .main import install_program


# Functions
def download_executable(download_urls: dict[str, str], program_name: str, append_to_path: str = "") -> bool:
	""" Ask the user if they want to download the program (ex: waifu2x-ncnn-vulkan).
	If yes, try to download the program from the GitHub releases page.

	Args:
		download_urls  (dict[str, str]):  The URLs to download the program from.
		program_name   (str):             The name of the program to download.

	Returns:
		bool: True if the program is now ready to use, False otherwise.
	"""
	# Ask the user if they want to download the upscaler
	program_url: str = next(iter(download_urls.values())).split("/download/")[0]
	warning(
		f"Program executable not found, would you like to download it automatically from GitHub? (Y/n) :\n"
		f"({program_url})"
	)

	# Handle the user's response
	if input().lower() == "n":
		info("User declined to download the upscaler.")
		return False

	# Get the platform
	system: str = platform.system()
	download_url: str = download_urls.get(system, "")
	if not download_url:
		warning(
			f"Unsupported platform: {system}, please download the program manually from the following URL:\n"
			f"  {program_url}"
		)
		return False

	# Download the upscaler
	if not install_program(download_url, program_name=program_name, append_to_path=append_to_path):
		warning("Failed to download the upscaler, please download it manually from the following URL:")
		print(f"  {download_url}")
		return False

	return True

def check_executable(
	executable: str,
	executable_help_text: str,
	download_urls: dict[str, str],
	append_to_path: str = ""
) -> None:
	""" Check if the executable exists, optionally download it if it doesn't.

	Args:
		executable            (str):             The path to the executable.
		executable_help_text  (str):             The help text to check for in the executable's output.
		download_urls         (dict[str, str]):  The URLs to download the executable from.
		append_to_path        (str):             The path to append to the executable's path.
			(ex: "bin" if executables are in the bin folder)
	"""
	program_name: str = os.path.basename(executable)
	try_download: bool = True

	# Run the command, capture output, don't check exit code immediately
	try:
		result: subprocess.CompletedProcess[str] = subprocess.run(
			[executable, "-h"],
			capture_output=True,
			text=True,  # Decode stdout/stderr as text
			check=False # Don't raise exception on non-zero exit code
		)

		# If the command failed (no help text matching), try to download the upscaler
		try_download: bool = executable_help_text.lower() not in result.stdout.lower()
	except FileNotFoundError:
		try_download: bool = True

	# If the command failed, try to download the upscaler
	if try_download:
		if not download_executable(download_urls, program_name, append_to_path=append_to_path):
			warning(f"'{program_name}' is required but not available. Exiting.")
		else:
			info(f"'{program_name}' downloaded successfully, please restart the script.")
		sys.exit(1)




""" A collection of utility modules designed to simplify and enhance the development process.

This package provides various tools and utilities for common development tasks including:

Key Features:
- Continuous delivery utilities (GitHub, PyPI)
- Display and logging utilities (print)
- File and I/O management (io)
- Decorators for common patterns
- Context managers
- Archive and backup tools
- Parallel processing helpers
- Collection utilities
- Doctests utilities

"""
# ruff: noqa: F403

# Imports
from .all_doctests import *
from .archive import *
from .backup import *
from .collections import *

# Folders
from .continuous_delivery import *
from .ctx import *
from .decorators import *
from .image import *
from .io import *
from .parallel import *
from .print import *

# Version (handle case where the package is not installed)
import importlib.metadata

try:
	__version__: str = importlib.metadata.version("stouputils")
except importlib.metadata.PackageNotFoundError:
	__version__: str = "0.0.0-dev"






# pyright: reportUnusedImport=false
# ruff: noqa: F401

# Imports
from typing import Any

import cv2
import numpy as np
from numpy.typing import NDArray


# Functions
def check_image(image: NDArray[Any], ignore_dtype: bool = False) -> None:
	""" Check if the image is valid

	Args:
		image         (NDArray[Any]):  Image to check
		ignore_dtype  (bool):        Ignore the dtype check
	"""
	# Check input data
	assert isinstance(image, np.ndarray), "Image must be a numpy array"
	assert len(image.shape) >= 2, "Image must have at least 2 dimensions"

	# Check dtype
	if not ignore_dtype:
		dtypes: list[type] = [np.uint8, np.int16, np.float32, np.float64]
		assert image.dtype in dtypes, f"Image must be of type {dtypes}"





# pyright: reportUnknownMemberType=false
# pyright: reportUnknownArgumentType=false
# pyright: reportUnknownVariableType=false

# Imports
import SimpleITK as Sitk

from .common import Any, NDArray, check_image, np


# Functions
def bias_field_correction_image(image: NDArray[Any], ignore_dtype: bool = False) -> NDArray[Any]:
	""" Apply a bias field correction to an image. (N4 Filter)

	Args:
		image         (NDArray[Any]):  Image to apply the bias field correction (can't be 8-bit unsigned integer)
		ignore_dtype  (bool):        Ignore the dtype check
	Returns:
		NDArray[Any]: Image with the curvature flow filter applied

	>>> ## Basic tests
	>>> image = np.random.randint(0, 255, size=(10,10), dtype=np.uint8) / 255
	>>> corrected = bias_field_correction_image(image)
	>>> corrected.shape == image.shape
	True
	>>> corrected.dtype == np.float64
	True

	>>> ## Test invalid inputs
	>>> bias_field_correction_image("not an image")
	Traceback (most recent call last):
		...
	AssertionError: Image must be a numpy array
	"""
	# Check input data
	check_image(image, ignore_dtype=ignore_dtype)

	# If the image is 3D, convert to grayscale first
	if image.ndim == 3:
		image = np.mean(image, axis=-1)

	# Convert numpy array to SimpleITK image
	image_sitk: Sitk.Image = Sitk.GetImageFromArray(image)

	# Create binary mask of the head region
	transformed: Sitk.Image = Sitk.RescaleIntensity(image_sitk)  # Normalize intensities
	transformed = Sitk.LiThreshold(transformed, 0, 1)  # Apply Li thresholding
	head_mask: Sitk.Image = transformed

	# Downsample images to speed up bias field estimation
	shrink_factor: int = 4  # Reduce image size by factor of 4
	input_image: Sitk.Image = Sitk.Shrink(
		image_sitk,
		[shrink_factor] * image_sitk.GetDimension()  # Apply shrink factor to all dimensions
	)
	mask_image: Sitk.Image = Sitk.Shrink(
		head_mask,
		[shrink_factor] * image_sitk.GetDimension()
	)

	# Apply N4 bias field correction
	corrector = Sitk.N4BiasFieldCorrectionImageFilter()
	corrector.Execute(input_image, mask_image)

	# Get estimated bias field and apply correction
	log_bias_field: Sitk.Image = Sitk.Cast(
		corrector.GetLogBiasFieldAsImage(image_sitk), Sitk.sitkFloat64
	)
	corrected_image_full_resolution: Sitk.Image = image_sitk / Sitk.Exp(log_bias_field)

	# Convert back to numpy array and return
	return Sitk.GetArrayFromImage(corrected_image_full_resolution)





# Imports
from .common import Any, NDArray, check_image, cv2, np


# Functions
def translate_image(image: NDArray[Any], x: float, y: float, padding: int = 0, ignore_dtype: bool = False) -> NDArray[Any]:
	""" Translate an image

	Args:
		image         (NDArray[Any]):  Image to translate
		x             (float):       Translation along the x axis (between -1 and 1)
		y             (float):       Translation along the y axis (between -1 and 1)
		padding       (int):         Padding that has been added to the image before calling this function
		ignore_dtype  (bool):        Ignore the dtype check
	Returns:
		NDArray[Any]: Translated image

	>>> ## Basic tests
	>>> image = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]).astype(np.uint8)
	>>> translate_image(image, 0.5, 0.5).tolist()
	[[0, 0, 0], [0, 1, 2], [0, 4, 5]]

	>>> translate_image(image, 0, -2/3).tolist()
	[[7, 8, 9], [0, 0, 0], [0, 0, 0]]

	>>> ## Test invalid inputs
	>>> translate_image(image, 2, 0)
	Traceback (most recent call last):
		...
	AssertionError: x must be between -1 and 1, got 2

	>>> translate_image(image, 0, 2)
	Traceback (most recent call last):
		...
	AssertionError: y must be between -1 and 1, got 2

	>>> translate_image("not an image", 0, 0)
	Traceback (most recent call last):
		...
	AssertionError: Image must be a numpy array

	>>> translate_image(image, 0, 0, padding=-1)
	Traceback (most recent call last):
		...
	AssertionError: padding must be positive, got -1
	"""
	# Check input data
	check_image(image, ignore_dtype=ignore_dtype)
	assert isinstance(x, float | int), f"x must be a number, got {type(x)}"
	assert isinstance(y, float | int), f"y must be a number, got {type(y)}"
	assert -1 <= x <= 1, f"x must be between -1 and 1, got {x}"
	assert -1 <= y <= 1, f"y must be between -1 and 1, got {y}"
	assert isinstance(padding, int), f"padding must be an integer, got {type(padding)}"
	assert padding >= 0, f"padding must be positive, got {padding}"

	# Get image dimensions
	height, width = image.shape[:2]
	original_width: int = width - 2 * padding
	original_height: int = height - 2 * padding

	# Convert relative translations to absolute pixels
	x_pixels: int = int(x * original_width)
	y_pixels: int = int(y * original_height)

	# Create translation matrix
	translation_matrix: NDArray[Any] = np.array([[1, 0, x_pixels], [0, 1, y_pixels]], dtype=np.float32)

	# Apply affine transformation
	return cv2.warpAffine(image, translation_matrix, (width, height))

